from enum import Enum
from logger import Logger
from typing import Dict, List, Any, Union, Optional
from aiohttp import web
import json
import os
import time


"""
Формат тест кейса таков:
{
  # ответа от сервера. строка, словарь или null
  "response": Union[str|Dict[str, Any]|null],

  # метаданные ответа(могут отсутствовать и тогда будут использованы значения по умолчанию)
  "responseMeta": {
    # строковое перечисление типа возвращаемых данных(по умолчанию dtext)
    "datatype": Enum(str) {dtext, djson, dxml},
         
    # строковое перечисление источника данных для ответа на запрос(по умолчанию stext)
    # если значение sfile, то считаем что данные ответа лежат в файле по пути:
    #    "/linuxcash/cash/servicestub/services/<имя_сервиса>/ + <значение_поля_response>"
    "source": Enum(str) {stext, sfile}
  },

  # возвращаемый статус http(если присутствует, то сервис вернёт его на запрос)
  "status": {
    # код статуса выполнения
    "code": int
  },

  # возвращаемая ошибка http(если присутствует, то сервис вернёт её на запрос)
  "error": {
    # код ошибки
    "code": int
  },

  # заполняемые заголовки в ответе от сервера, задаются в виде словаря. Могут отсутствовать
  "headers": Dict[str, Any]

  # счётчик ответов на запрос. Если отсутствует, то ответ будет возвращаться всегда
  "caseCount": int,

  # таймаут перед ответом сервиса. Может отсутствовать
  "timeout": float

  # комментарий относительно кейса. Может отсутствовать
  "comment": str
}

Кейсы могут быть заданы списком для пути
{
  "<url-путь>": [
    <case1>,
    <case2>,
    <case3>
  ]
}
или, если нужны какие-то параметры - словарём со списками
{
  "<url-путь>": {
    "набор параметров 1": [
      <case1>,
      <case2>,
      <case3>
    ],
    "набор параметров 2": [
      <case1>,
      <case2>,
      <case3>
    ]
  }
}
"""


class EDataType(str, Enum):
    dtext = 'text'
    djson = 'json'
    dxml = 'xml'


class ESource(str, Enum):
    stext = 'text'
    sfile = 'file'


class ResponseMeta:
    datatype: Enum
    source: Enum

    def __init__(self, datatype: EDataType=EDataType.dtext, source: ESource=ESource.stext, **kwargs):
        self.datatype = datatype
        self.source = source


class Status:
    code: int

    def __init__(self, code: int=200, **kwargs):
        self.code = code


#Такие сложные махинации нужны потому, что в aiohttp нельзя вернуть произвольную ошибку
#Так же нелья возвращать одну и ту же (один и тот же объект)
class Error:
    code: int

    class HttpError(web.HTTPException):
        def __init__(self, code: int=200):
            __class__.status_code = code
            super().__init__()

    def __init__(self, code: int=200, **kwargs):
        self.code = code

    def makeHttpError(self) -> HttpError:
        """ Формирование объекта ошибки """
        return __class__.HttpError(self.code)


class Headers:
    headers: Dict[str, Any]

    def __init__(self, headers: Dict[str, Any], **kwargs):
        self.headers = headers


class Case:
    response: Any
    responseMeta: ResponseMeta
    status: Union[Status, None]
    error: Union[Error, None]
    headers: Union[Headers, None]
    caseCount: Union[int, None]
    timeout: Union[float, None]
    comment: Union[str, None]

    def __init__(self, response: Any=None, responseMeta: Optional[Dict[str, Any]]=None,
                status: Optional[Dict[str, Any]]=None, error: Optional[Dict[str, Any]]=None,
                headers: Optional[Dict[str, Any]]=None, caseCount: Optional[int]=None,
                timeout: Optional[float]=None, comment: Optional[str]=None, **kwargs):
        self.response = response
        self.responseMeta = ResponseMeta() if responseMeta is None else ResponseMeta(**responseMeta)
        self.status = None if status is None else Status(**status)
        self.error = None if error is None else Error(**error)
        self.headers = None if headers is None else Headers(headers)
        self.caseCount = caseCount
        self.timeout = timeout
        self.comment = comment

    def createResponse(self, serviceDir: str) -> Union[web.Response, Error.HttpError]:
        """ Формирование ответа

        Args:
            serviceDir: папка с сервисом
        """
        if self.error is not None:
            return self.error.makeHttpError()

        # первое - это получение данных
        content = self.response
        if self.responseMeta.source == ESource.sfile:
            with open(os.path.join(serviceDir, self.response)) as f:
                content = f.read()

        # второе - это формирование ответа
        status = 200 if self.status is None else self.status.code
        headers = None if self.headers is None else self.headers.headers
        if self.responseMeta.datatype == EDataType.djson:
            if isinstance(content, str):
                return web.json_response(text=content, status=status, headers=headers)
            return web.json_response(content, status=status, headers=headers)
        elif self.responseMeta.datatype == EDataType.dxml:
            return web.Response(text=content, status=status, content_type="application/xml", headers=headers)
        else:
            return web.Response(text=content, status=status, headers=headers)


class CaseFile:
    path: str
    time: Union[float, None]

    def __init__(self, path: str):
        self.path = path
        self.time = None


class BasicService(object):
    log: Logger
    servicePath: str  # url-путь сервиса
    serviceDir: str  # путь до директории с сервисом(<имя_сервиса>.py)
    cases: Dict[str, Dict[str, List[Case]]]
    caseFiles: List[CaseFile]

    def __init__(self, service: str):
        self.log = Logger()
        self.servicePath = ''
        self.cases = {}
        self.caseFiles = []
        self.serviceDir = os.path.dirname(os.path.abspath(service))

    async def _basicHandler(self, request: web.Request) -> web.Response:
        """ Базовый обрабоотчик запросов. Просто возвращает ответ для пути запроса

        Args:
            request: запрос
        """
        return self._createResponse(request.path)

    def __loadCases(self, caseFile: CaseFile) -> None:
        """ Обновление ответов и метаинформации из указанного файла

        Args:
            caseFile: файл для загрузки
        """
        try:
            with open(caseFile.path, 'r') as file:
                cases = json.load(file)
            caseFile.time = os.path.getmtime(caseFile.path)
            for path, casesValue in cases.items():
                if isinstance(casesValue, dict):
                    self.cases[path] = {}
                    for params, caseList in casesValue.items():
                        self.cases[path][params] = [Case(**case) for case in caseList]
                else:
                    self.cases[path] = {'': [Case(**case) for case in casesValue]}
        except Exception as e:
            self.log.error(str(e))

    def _loadCases(self, filename: str) -> None:
        """ Загрузка ответов из указанного файла

        Args:
            filename: файл для загрузки
        """
        filename = os.path.join(self.serviceDir, filename)

        for caseFile in self.caseFiles:
            if caseFile.path == filename:
                self.__loadCases(caseFile)
                return

        caseFile = CaseFile(filename)
        self.__loadCases(caseFile)
        self.caseFiles.append(caseFile)

    def _loadAllCases(self) -> None:
        """ Загрузка ответов из всех файлов с расширением '.json' в serviceDir """
        self.cases.clear()
        self.caseFiles.clear()
        for file in sorted(os.listdir(self.serviceDir)):
            if file.endswith(".json"):
                self._loadCases(file)

    def __convertMethodName(self, methodName: str) -> str:
        """ Приведение имени метода к универсальному виду

        Args:
            methodName: имя метода
        """
        if self.servicePath and self.servicePath != methodName and self.servicePath in methodName:
            return methodName.removeprefix(self.servicePath).removeprefix('/')
        return methodName

    def _hasCases(self, methodName: str, params: str='') -> bool:
        """ Проверка на наличие ответов

        Args:
            methodName: имя метода
            params: параметры метода
        """
        methodName = self.__convertMethodName(methodName)

        for caseFile in self.caseFiles:
            time = os.path.getmtime(caseFile.path)
            if time != caseFile.time:
                self.cases.clear()
                for caseFile in self.caseFiles:
                    self.__loadCases(caseFile)
                break

        if methodName not in self.cases:
            return False
        return params in self.cases[methodName]

    def _createResponse(self, methodName: str, params: str='') -> web.Response:
        """ Формирование ответа

        Args:
            methodName: имя метода
            params: параметры метода
        """
        methodName = self.__convertMethodName(methodName)

        if not self._hasCases(methodName, params):
            message = f'Не найден метод "{methodName}"'
            if params:
                message = message + f' с параметрами "{params}"'
            raise web.HTTPBadRequest(text=message)

        if not len(self.cases[methodName][params]):
            message = f'В кейсах не осталось значений для {methodName}'
            if params:
                message = message + f' ({params})'
            self.log.warn(message)
            raise web.HTTPInternalServerError(text=message)

        case = self.cases[methodName][params][0]

        if case.timeout is not None:
            time.sleep(case.timeout)

        response = case.createResponse(self.serviceDir)

        # уменьшить счётчик кейса, если он есть
        if case.caseCount is not None:
            case.caseCount -= 1
            if case.caseCount == 0:
                del self.cases[methodName][params][0]

        if case.comment is not None:
            self.log.info(f"Используется ответ: '{case.comment}'")

        if isinstance(response, Error.HttpError):
            raise response
        return response
